from logging import getLogger
from shlex import quote
from typing import Dict

from airflow.models.dag import DAG
from airflow.models.operator import Operator
from airflow.providers.ssh.operators.ssh import SSHOperator
from airflow_balancer import Host, Port
from airflow_pydantic import SSHOperatorArgs
from supervisor_pydantic.convenience import SupervisorTaskStep

from airflow_supervisor.config import SupervisorSSHAirflowConfiguration

from .local import Supervisor

__all__ = ("SupervisorSSH",)

_log = getLogger(__name__)


class SupervisorSSH(Supervisor):
    # Mimic SSH Operator: https://airflow.apache.org/docs/apache-airflow-providers-ssh/stable/_api/airflow/providers/ssh/operators/ssh/index.html
    def __init__(
        self,
        dag: DAG,
        cfg: SupervisorSSHAirflowConfiguration,
        host: "Host" = None,
        port: "Port" = None,
        **kwargs,
    ):
        if isinstance(cfg, dict):
            # NOTE: used in airflow-pydantic rendering
            cfg = SupervisorSSHAirflowConfiguration.model_validate(cfg)

        for attr in ("command_prefix",):
            if attr in kwargs:
                _log.info(f"Setting {attr} to {kwargs.get(attr)}")
                setattr(self, f"_{attr}", kwargs.pop(attr))
            elif cfg and getattr(cfg, attr):
                _log.info(f"Setting {attr} to {getattr(cfg, attr)}")
                setattr(self, f"_{attr}", getattr(cfg, attr))
            else:
                _log.info(f"Setting {attr} to empty string")
                setattr(self, f"_{attr}", "")

        self._ssh_operator_kwargs = cfg.ssh_operator_args.model_dump(exclude_none=True)

        pydantic_fields = (
            cfg.ssh_operator_args.__pydantic_fields__.keys()
            if hasattr(cfg.ssh_operator_args, "__pydantic_fields__")
            else cfg.ssh_operator_args.__fields__
        )
        for attr in pydantic_fields:
            if attr in kwargs:
                _log.info(f"Setting {attr} to {kwargs.get(attr)}")
                self._ssh_operator_kwargs[attr] = kwargs.pop(attr)
                setattr(cfg, attr, self._ssh_operator_kwargs[attr])

        # Integrate with airflow-balancer, use host if provided
        if host:
            _log.info(f"Setting host to {host.name}")
            self._ssh_operator_kwargs["remote_host"] = host.name
            self._ssh_operator_kwargs["ssh_hook"] = host.hook()
            cfg.ssh_operator_args = SSHOperatorArgs(**self._ssh_operator_kwargs)

            # Ensure host matches the configuration
            cfg.convenience.host = host.name

            # Extract pool if available
            if host.pool and not cfg.airflow.pool:
                _log.info(f"Setting airflow pool to {host.pool}")
                cfg.airflow.pool = host.pool

        if port:
            _log.info(f"Setting port to {port.port}")
            # Ensure port matches the configuration
            cfg.convenience.port = f"*:{port.port}"

        if host or port:
            # revalidate
            _log.info("Revalidating configuration")
            cfg._setup_convenience_defaults()

        super().__init__(dag=dag, cfg=cfg, **kwargs)

    def get_step_kwargs(self, step: SupervisorTaskStep) -> Dict:
        if step == "configure-supervisor":
            # TODO
            return {
                **self._ssh_operator_kwargs,
                "command": f"""
{self._command_prefix}
_supervisor_convenience {step} '{self._cfg.model_dump_json()}'
""",
            }
        elif step in ("start-supervisor", "stop-supervisor", "unconfigure-supervisor", "force-kill"):
            # must be done via SSH
            return {
                **self._ssh_operator_kwargs,
                "command": f"""
{self._command_prefix}
_supervisor_convenience {step} --cfg {quote(str(self._cfg._pydantic_path))}
""",
            }
        else:
            # can be done via XMLRPC API
            return super().get_step_kwargs(step=step)

    def get_step_operator(self, step: SupervisorTaskStep) -> Operator:
        if step in (
            "configure-supervisor",
            "start-supervisor",
            "stop-supervisor",
            "unconfigure-supervisor",
            "force-kill",
        ):
            return SSHOperator(
                **{
                    "task_id": f"{self._dag.dag_id}-{step}",
                    **self.get_base_operator_kwargs(),
                    **self.get_step_kwargs(step),
                }
            )
        else:
            # can be done via XMLRPC API
            return super().get_step_operator(step=step)



from airflow.models.dag import DAG

from airflow_supervisor import Supervisor, SupervisorAirflowConfiguration


class TestDag:
    def test_instantiation(self, supervisor_airflow_configuration: SupervisorAirflowConfiguration):
        dag = DAG(dag_id="test_dag", default_args={}, schedule=None, params={})
        s = Supervisor(dag=dag, cfg=supervisor_airflow_configuration)
        assert len(dag.tasks) == 17
        assert dag.catchup is False
        assert dag.concurrency == 1
        assert dag.max_active_tasks == 1
        assert dag.max_active_runs == 1

        assert s.configure_supervisor in dag.tasks
        assert s.configure_supervisor in dag.tasks
        assert s.start_supervisor in dag.tasks
        assert s.start_programs in dag.tasks
        assert s.check_programs in dag.tasks
        assert s.restart_programs in dag.tasks
        assert s.stop_programs in dag.tasks
        assert s.stop_supervisor in dag.tasks
        assert s.unconfigure_supervisor in dag.tasks

        for task in dag.tasks:
            assert task.dag == dag, f"Task {task.task_id} is not associated with the DAG"
            assert task.pool == supervisor_airflow_configuration.airflow.pool, (
                f"Task {task.task_id} does not have the correct pool set"
            )



from typing import Optional, Type

from airflow_balancer import Host, Port
from airflow_pydantic import ImportPath, Task, TaskArgs
from pydantic import Field, field_validator

from .supervisor import SupervisorAirflowConfiguration, SupervisorSSHAirflowConfiguration

__all__ = (
    "SupervisorOperatorArgs",
    "SupervisorTaskArgs",
    "SupervisorSSHOperatorArgs",
    "SupervisorSSHTaskArgs",
    "SupervisorOperator",
    "SupervisorTask",
    "SupervisorSSHOperator",
    "SupervisorSSHTask",
)


class SupervisorTaskArgs(TaskArgs, extra="allow"):
    cfg: SupervisorAirflowConfiguration


# Alias
SupervisorOperatorArgs = SupervisorTaskArgs


class SupervisorTask(Task, SupervisorTaskArgs):
    operator: ImportPath = Field(default="airflow_supervisor.Supervisor", validate_default=True)

    @field_validator("operator")
    @classmethod
    def validate_operator(cls, v: Type) -> Type:
        if not isinstance(v, Type) and issubclass(v, SupervisorAirflowConfiguration):
            raise ValueError(f"operator must be 'airflow_supervisor.Supervisor', got: {v}")
        return v


# Alias
SupervisorOperator = SupervisorTask


class SupervisorSSHTaskArgs(TaskArgs, extra="allow"):
    cfg: SupervisorAirflowConfiguration
    host: Optional[Host] = Field(
        default=None,
        description="The host to connect to for SSH, if not otherwise provided in configs",
    )
    port: Optional[Port] = Field(
        default=None,
        description="The port to user for Supervisor, if not otherwise provided in configs",
    )


# Alias
SupervisorSSHOperatorArgs = SupervisorSSHTaskArgs


class SupervisorSSHTask(Task, SupervisorSSHTaskArgs):
    operator: ImportPath = Field(default="airflow_supervisor.SupervisorSSH", validate_default=True)

    @field_validator("operator")
    @classmethod
    def validate_operator(cls, v: Type) -> Type:
        if not isinstance(v, Type) and issubclass(v, SupervisorSSHAirflowConfiguration):
            raise ValueError(f"operator must be 'airflow_supervisor.SupervisorSSH', got: {v}")
        return v


# Alias
SupervisorSSHOperator = SupervisorSSHTask



from typing import Optional

from airflow_pydantic import SSHOperatorArgs
from pydantic import Field
from supervisor_pydantic import SupervisorConvenienceConfiguration

from .airflow import AirflowConfiguration

__all__ = (
    "SupervisorAirflowConfiguration",
    "SupervisorSSHAirflowConfiguration",
    "SSHOperatorArgs",
    "load_airflow_config",
    "load_airflow_ssh_config",
)


class SupervisorAirflowConfiguration(SupervisorConvenienceConfiguration):
    airflow: AirflowConfiguration = Field(
        default_factory=AirflowConfiguration, description="Required options for airflow integration"
    )

    # Should the programs be stopped when the DAG finishes?
    stop_on_exit: Optional[bool] = Field(default=True, description="Stop supervisor on dag completion")
    # Should the supervisor folder be removed on dag completion?
    cleanup: Optional[bool] = Field(
        default=True, description="Cleanup supervisor folder on dag completion. Note: stop_on_exit must be True"
    )
    # For Jobs that do not shutdown, e.g. stop_on_exit=False, one might want to configure them to
    # restart when the DAG is rescheduled by airflow or retriggered by airflow-ha
    restart_on_initial: Optional[bool] = Field(
        default=False,
        description="Restart the job when the DAG is run directly via airflow (NOT retriggered). This is useful for jobs that do not shutdown",
    )
    restart_on_retrigger: Optional[bool] = Field(
        default=False,
        description="Restart the job when the DAG is retriggered. This is useful for jobs that do not shutdown",
    )


class SupervisorSSHAirflowConfiguration(SupervisorAirflowConfiguration):
    command_prefix: Optional[str] = Field(default="")

    ssh_operator_args: Optional[SSHOperatorArgs] = Field(
        default_factory=SSHOperatorArgs,
        description="SSH Operator arguments to use for remote execution.",
    )


load_airflow_config = SupervisorAirflowConfiguration.load
load_airflow_ssh_config = SupervisorSSHAirflowConfiguration.load
